/*
 * The aim is to get a simpler inteface to the database routines.
 * All the tidieous messing around with tuples is supposed to be hidden
 * by this function.
 */
/* Author: Linus Tolke 
   (actually most if the code is "borrowed" from the distribution and just
    slightly modified)
 */

/*
 * Example program using this should look like this:

#include "c.h"

main()
{
    find_out_where_db_is();

    CPQconnect(dbname);

    {
    	struct { int len, char arr[20] } name;
	int age;

	if (CPQdo(24, "select lastname, salary from table1 where name = %s;",
	          et_varchar, &name, 0, 20,
		  et_EOIT
	          et_varchar, &name, 0, 20,
		  et_int, &age, 0, sizeof(int),
		  et_EORT) == true)
	{
	    do_something_with(name, age);
	}
	else
	{
	    We got some error, either to few hits or to many hits or something
	    completely different.
	}
    }

    CPQfinish();
}


Notice:
- When we exit the PQdo we have already allocated memory for the result,
  copied the result to the newly allocated area and freed the portal to 
  the database.

BUGS:
- We should be able to set the malloc function of our choise.
- Inserting strings should do quoting.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>

#include "simple.h"
#include "type.h"
#include "sqlca.h"
#include "libpq-fe.h"
#include "libpq/pqcomm.h"

static PGconn *	simple_connection;
static int	simple_debug = 0;
static int	committed = true;

static
register_error(int code, char *fmt, ...)
{
    va_list args;

    sqlca.sqlcode = code;
    va_start (args, fmt);
    vsprintf (sqlca.sqlerrm.sqlerrmc, fmt, args);
    va_end (args);
    sqlca.sqlerrm.sqlerrml = strlen (sqlca.sqlerrm.sqlerrmc);
}

/* This function returns a newly malloced string that has the ' and \
   in the argument quoted with \.
 */
static
char *
quote_postgres(char * arg)
{
    char * res = (char *)malloc(2 * strlen(arg) + 1);
    int i, ri;

    for (i = 0, ri = 0; arg[i]; i++, ri++)
    {
        switch (arg[i])
	{
	case '\'':
	case '\\':
	    res[ri++] = '\\';
	default:
	    ;
	}

	res[ri] = arg[i];
    }
    res[ri] = '\0';

    return res;
}


bool
CPQdo(int lineno, char * query, ...)
{
    va_list ap;
    bool status = false;
    char * copiedquery;
    PGresult *	results;
    PGnotify *	notify;
    enum Etype type;

    va_start(ap, query);

    sqlca.sqlcode = 0;
    copiedquery = strdup(query);

    type = va_arg(ap, enum Etype);

    /*
     * Now, if the type is one of the fill in types then we take the argument
     * and enter that in the string at the first %s position. Then if there
     * are any more fill in types we fill in at the next and so on.
     */
    while (type != et_EOIT)
    {
	void * value = NULL;
	short varcharsize;
	short size;
	short arrsize;

        char * newcopy;
	char * mallocedval = NULL;
	char * tobeinserted = NULL;
	char * p;
	char buff[20];

	/* Some special treatment is needed for records since we want their
	   contents to arrive in a comma-separated list on insert (I think). */

	value = va_arg(ap, void *);
	varcharsize = va_arg(ap, short);
	size = va_arg(ap, short);
	arrsize = va_arg(ap, short);

	switch (type) {
	case et_char:
	case et_short:
	case et_int:
	    sprintf(buff, "%d", *(int*)value);
	    tobeinserted = buff;
	    break;

	case et_unsigned_char:
	case et_unsigned_short:
	case et_unsigned_int:
	    sprintf(buff, "%d", *(unsigned int*)value);
	    tobeinserted = buff;
	    break;

	case et_long:
	    sprintf(buff, "%ld", *(long*)value);
	    tobeinserted = buff;
	    break;

	case et_unsigned_long:
	    sprintf(buff, "%ld", *(unsigned long*)value);
	    tobeinserted = buff;
	    break;

	case et_float:
	    sprintf(buff, "%.14g", *(float*)value);
	    tobeinserted = buff;
	    break;

	case et_double:
	    sprintf(buff, "%.14g", *(double*)value);
	    tobeinserted = buff;
	    break;

	case et_varchar:
	case et_varchar2:
	    {
		struct generic_varchar * var = (struct generic_varchar*)value;

		newcopy = (char *)malloc(var->len + 1);
		strncpy(newcopy, var->arr, var->len);
		newcopy[var->len] = '\0';
	    
		mallocedval = (char *)malloc(2 * strlen(newcopy) + 3);
		strcpy(mallocedval, "'");
		strcat(mallocedval, quote_postgres(newcopy));
		strcat(mallocedval, "'");

		free(newcopy);

		tobeinserted = mallocedval;
	    }
	    break;

	default:
	    /* Not implemented yet */
	    register_error(-1, "Unsupported type %s on line %d.\n",
			   type_name(type), lineno);
	    return false;
	    break;
	}

	/* Now tobeinserted points to an area that is to be inserted at 
	   the first %s 
	 */
	newcopy = (char *)malloc(strlen(copiedquery) 
				 + strlen(tobeinserted) 
				 + 1);
	strcpy(newcopy, copiedquery);
	if ((p = strstr(newcopy, ";;")) == NULL)
	{
	    /* We have an argument but we dont have the matched up string 
	       in the string
	     */
	    register_error(-1, "Too many arguments line %d.\n", lineno);
	    return false;
	}
	else
	{
	    strcpy(p, tobeinserted);
	    /* The strange thing in the second argument is the rest of the
	       string from the old string */
	    strcat(newcopy, 
		   copiedquery
		   + ( p - newcopy ) 
		   + 2 /* Length of ;; */);
	}

	/* Now everything is safely copied to the newcopy. Lets free the
	   oldcopy and let the copiedquery get the value from the newcopy.
	 */
	if (mallocedval != NULL)
	{
	    free(mallocedval);
	    mallocedval = NULL;
	}

	free(copiedquery);
	copiedquery = newcopy;

        type = va_arg(ap, enum Etype);
    }

    /* Check if there are unmatched things left. */
    if (strstr(copiedquery, ";;") != NULL)
    {
	register_error(-1, "Too few arguments line %d.\n", lineno);
	return false;
    }

    /* Now then request is built. */

    if (committed)
    {
	if ((results = PQexec (simple_connection, "begin")) == NULL) {
	    register_error(-1, "Error starting transaction line %d", lineno);
	    return false;
	}
	PQclear (results);
	committed = 0;
    }

    CPQlog("CPQdo: QUERY: %s\n", copiedquery);
    results = PQexec(simple_connection, copiedquery); 
    free(copiedquery);

    if (results == NULL)
    {
        CPQlog("CPQdo: error: %s", PQerrorMessage(simple_connection));
	register_error(-1, "Postgres error: %s line %d", 
		       PQerrorMessage(simple_connection), lineno);
    }
    else switch(PQresultStatus(results)) 
    {
        int m,n,x;

    case PGRES_TUPLES_OK:
	/* XXX Cheap Hack. For now, we see only the last group
	 * of tuples.  This is clearly not the right
	 * way to do things !!
	 */

	m = PQnfields(results);
	n = PQntuples(results);
    
	if (n < 1)
	{
	    CPQlog("CPQdo: Incorrect number of matches: %d\n", n);
	    register_error(1, "Data not found line %d.", lineno);
	    break;
	}

	if (n > 1)
	{
	    CPQlog("CPQdo: Incorrect number of matches: %d\n", n);
	    register_error(-1, "To many matches line %d.", lineno);
	    break;
	}

	status = true;

	for (x = 0; x < m && status; x++) 
	{
	    void * value = NULL;
	    short varcharsize;
	    short size;
	    short arrsize;

	    char 	*pval = PQgetvalue(results,0,x);
	    long int	* res_int;
	    char 	** res_charstar;
	    char	* res_char;
	    int		res_len;
	    char	* scan_length;

	    CPQlog("CPQdo: RESULT: %s\n", pval ? pval : "");

	    /* No the pval is a pointer to the value. */
	    /* We will have to decode the value */
	    type = va_arg(ap, enum Etype);
	    value = va_arg(ap, void *);
	    varcharsize = va_arg(ap, short);
	    size = va_arg(ap, short);
	    arrsize = va_arg(ap, short);

	    switch (type)
	    {
		long res;
		unsigned long ures;
		double dres;

	    case et_char:
	    case et_short:
	    case et_int:
	    case et_long:
		if (pval)
		{
		    res = strtol(pval, &scan_length, 10);
		    if (*scan_length != '\0') /* Garbage left */
		    {
			register_error(-1, "Not correctly formatted int type: %s line %d.",
				       pval, lineno);
			status = false;
			res = 0L;
		    }
		}
		else
		    res = 0L;

		/* Again?! Yes */
		switch (type)
		{
		case et_char:
		    *(char *)value = (char)res;
		    break;
		case et_short:
		    *(short *)value = (short)res;
		    break;
		case et_int:
		    *(int *)value = (int)res;
		    break;
		case et_long:
		    *(long *)value = res;
		    break;
		default:
		    /* Cannot happen */
		    break;
		}
		break;

	    case et_unsigned_char:
	    case et_unsigned_short:
	    case et_unsigned_int:
	    case et_unsigned_long:
		if (pval)
		{
		    ures = strtoul(pval, &scan_length, 10);
		    if (*scan_length != '\0') /* Garbage left */
		    {
			register_error(-1, "Not correctly formatted unsigned type: %s line %d.",
				       pval, lineno);
			status = false;
			ures = 0L;
		    }
		}
		else
		    ures = 0L;

		/* Again?! Yes */
		switch (type)
		{
		case et_unsigned_char:
		    *(unsigned char *)value = (unsigned char)ures;
		    break;
		case et_unsigned_short:
		    *(unsigned short *)value = (unsigned short)ures;
		    break;
		case et_unsigned_int:
		    *(unsigned int *)value = (unsigned int)ures;
		    break;
		case et_unsigned_long:
		    *(unsigned long *)value = ures;
		    break;
		default:
		    /* Cannot happen */
		    break;
		}
		break;


	    case et_float:
	    case et_double:
		if (pval)
		{
		    dres = strtod(pval, &scan_length);
		    if (*scan_length != '\0') /* Garbage left */
		    {
			register_error(-1, "Not correctly formatted floating point type: %s line %d.",
				       pval, lineno);
			status = false;
			dres = 0.0;
		    }
		}
		else
		    dres = 0.0;

		/* Again?! Yes */
		switch (type)
		{
		case et_float:
		    *(float *)value = (float)res;
		    break;
		case et_double:
		    *(double *)value = res;
		    break;
		default:
		    /* Cannot happen */
		    break;
		}
		break;


	    case et_varchar:
		{
		    struct generic_varchar * var = 
			(struct generic_varchar*)value;

		    strncpy(var->arr, pval, varcharsize);
		    var->len = strlen(pval);
		    if (var->len > varcharsize)
			var->len = varcharsize;
		}
		break;

	    case et_EORT:
	        CPQlog("Too few arguments.\n");
		register_error(-1, "Too few arguments line %d.", lineno);
		status = false;
		break;

	    default:
		register_error(-1, "Unsupported type %s on line %d.\n",
			       type_name(type), lineno);
		return false;
		break;
	    }
	}

	type = va_arg(ap, enum Etype);

	if (status && type != et_EORT)
	{
	    register_error(-1, "Too many arguments line %d.\n", lineno);
	    return false;
	}

	PQclear(results);
	break;
    case PGRES_EMPTY_QUERY:
        /* do nothing */
	register_error(-1, "Empty query line %d.", lineno);
        break;
    case PGRES_COMMAND_OK:
        status = true;
        CPQlog("Ok: %s\n",PQcmdStatus(results));
	break;
    case PGRES_NONFATAL_ERROR:
    case PGRES_FATAL_ERROR:
    case PGRES_BAD_RESPONSE:
        CPQlog("Error: %s", PQerrorMessage(simple_connection));
	register_error(-1, "Error: %s line %d.", 
		       PQerrorMessage(simple_connection), lineno);
	break;
    case PGRES_COPY_OUT:
        CPQlog("Got PGRES_COPY_OUT ... tossing.\n");
        PQendcopy(results->conn);
	break;
    case PGRES_COPY_IN:
        CPQlog("Got PGRES_COPY_IN ... tossing.\n");
        PQendcopy(results->conn);
	break;
    default:
        CPQlog("Got something else, postgres error.\n");
	register_error(-1, "Postgres error line %d.", lineno);
	break;
    }

    /* check for asynchronous returns */
    notify = PQnotifies(simple_connection);
    if (notify) {
        CPQlog("ASYNC NOTIFY of '%s' from backend pid '%d' received\n",
	    notify->relname, notify->be_pid);
	free(notify);
    }

    va_end(ap);
    return status;
}


bool
CPQcommit(int lineno)
{
    PGresult *res;

    if ((res = PQexec (simple_connection, "end")) == NULL) {
	register_error(-1, "Error committing line %d", lineno);
	return;
    }
    PQclear (res);
    committed = 1;
}

bool
CPQrollback(int lineno)
{
    PGresult *res;

    if ((res = PQexec (simple_connection, "abort")) == NULL) {
	register_error(-1, "Error rolling back line %d", lineno);
	return;
    }
    PQclear (res);
    committed = 1;
}



bool
CPQsetdb(PGconn * newcon)
{
    CPQfinish();
    simple_connection = newcon;
    return true;
}

bool
CPQconnect(const char * dbname)
{
    char * name = strdup(dbname);
    CPQlog("CPQconnect: opening database %s\n", name);

    sqlca.sqlcode = 0;

    CPQsetdb(PQsetdb(NULL, NULL, NULL, NULL, name));

    free(name);
    name = NULL;

    if (PQstatus(simple_connection) == CONNECTION_BAD)
    {
        CPQfinish();
	CPQlog("CPQconnect: could not open database %s\n", dbname);
	register_error(-1, "CPQconnect: could not open database %s\n", dbname);
	return false;
    }
    return true;
}


bool
CPQstatus()
{
    return PQstatus(simple_connection) != CONNECTION_BAD;
}


bool
CPQfinish()
{
    if (simple_connection != NULL)
    {
        CPQlog("CPQfinish: finishing.\n");
        PQfinish(simple_connection);
    }
    else
        CPQlog("CPQfinish: called an extra time.\n");
    return true;
}

void
CPQdebug(int n)
{
    simple_debug = n;
    CPQlog("CPQdebug: set to %d\n", simple_debug);
}

void
CPQlog(const char * format, ...)
{
    va_list ap;
    if (simple_debug) 
    {
        char * f = (char *) malloc(strlen(format) + 20);

	sprintf(f, "CPQdo [%d]: %s", getpid(), format);

	va_start(ap, format);
        vfprintf(stderr, f, ap);
	va_end(ap);

	free(f);
    }
}
