/*
 * This file implements a data structure that is built and maintained by the
 * preprocessor.
 *
 * All types that can be handled for host variable declarations has to
 * be handled eventually.
 */

/*
 * Here are all the types that we are to handle. Note that it is the type
 * that is registered and that has nothing whatsoever to do with the storage
 * class.
 *
 * Simle types
 * integers: char, short, int, long (signed and unsigned)
 * floats: float, double
 *
 * Complex types:
 * VARCHAR, VARCHAR2 - Strings with length (maxlen is given in the declaration)
 * Arrays of simple types and of VARCHAR, VARCHAR2 (size given in declaration)
 * Records build of simple types, arrays and other records.
 *
 * Complicating things:
 * typedefs and struct names!
 *
 * Conclusion:
 * This is a typically recursive definition. A structure of typed list elements
 * would probably work fine:
 */
#include <stdio.h>

enum Etype {
    et_char = 1, et_unsigned_char, et_short, et_unsigned_short, 
    et_int, et_unsigned_int, et_long, et_unsigned_long,
    et_float, et_double,
    et_varchar, et_varchar2,
    et_array,
    et_record,
    et_EOIT,			/* End of insert types. */
    et_EORT			/* End of result types. */

};

#define IS_SIMPLE_TYPE(type) ((type) >= et_char && (type) <= et_varchar2)


struct type;
struct record_member {
    char *		name;
    struct type *	typ;
};
struct type {
    enum Etype		typ;
    unsigned short	size;	/* For array it is the number of elements.
				 * For varchar it is the maxsize of the area.
				 */
    union {
	struct type * element;	/* For an array this is the type of the 
				 * element */

	struct record_member ** members;
				/* A pointer to an array of members. */
    } u;
};

/* Everything is malloced. */
struct record_member * make_record_member(char *, struct type *);
struct type * make_simple_type(enum Etype);
struct type * make_varchar_type(enum Etype, unsigned short);
struct type * make_array_type(struct type *, unsigned short);
struct type * make_record_type(struct record_member *[]);

/* Frees a type. */
void free_record_member(struct record_member *);
void free_type(struct type *);

/* Dump a type.
   The type is dumped as:
   type-tag <comma> reference-to-variable <comma> arrsize <comma> size <comma>
   Where:
   type-tag is one of the simple types or varchar.
   reference-to-variable can be a reference to a struct element.
   arrsize is the size of the array in case of array fetches. Otherwise 0.
   size is the maxsize in case it is a varchar. Otherwise it is the size of 
       the variable (required to do array fetches of records).
 */
void dump_a_type(FILE *, const char * name, struct type *);

/* Returns a pointer to a string containing a simple type name. */
const char * type_name(enum Etype);

/* A simple struct to keep a variable and its type. */
struct temp_type {
    struct type * typ;
    const char * name;
};

/* A generic varchar type. */
struct generic_varchar {
    int len;
    char arr[1];
};
